@@ -1,10 +1,11 @@
 /*
  * ncheck.c --- given a list of inodes, generate a list of names
- * 
+ *
  * Copyright (C) 1994 Theodore Ts'o.  This file may be redistributed
  * under the terms of the GNU Public License.
  */
 
+#include "config.h"
 #include <stdio.h>
 #include <unistd.h>
 #include <stdlib.h>
@@ -18,18 +19,12 @@
 
 #include "debugfs.h"
 
-struct inode_info {
-	ext2_ino_t	ino;
-	ext2_ino_t	parent;
-	char		*pathname;
-};
-
 struct inode_walk_struct {
-	struct inode_info	*iarray;
+	ext2_ino_t		*iarray;
 	int			inodes_left;
 	int			num_inodes;
 	int			position;
-	ext2_ino_t		parent;
+	char			*parent;
 };
 
 static int ncheck_proc(struct ext2_dir_entry *dirent,
@@ -45,28 +40,27 @@ static int ncheck_proc(struct ext2_dir_entry *dirent,
 	if (iw->position <= 2)
 		return 0;
 	for (i=0; i < iw->num_inodes; i++) {
-		if (iw->iarray[i].ino == dirent->inode) {
-			iw->iarray[i].parent = iw->parent;
-			iw->inodes_left--;
+		if (iw->iarray[i] == dirent->inode) {
+			printf("%u\t%s/%.*s\n", iw->iarray[i], iw->parent,
+			       (dirent->name_len & 0xFF), dirent->name);
 		}
 	}
 	if (!iw->inodes_left)
 		return DIRENT_ABORT;
-	
+
 	return 0;
 }
 
 void do_ncheck(int argc, char **argv)
 {
 	struct inode_walk_struct iw;
-	struct inode_info	*iinfo;
 	int			i;
 	ext2_inode_scan		scan = 0;
 	ext2_ino_t		ino;
 	struct ext2_inode	inode;
 	errcode_t		retval;
 	char			*tmp;
-	
+
 	if (argc < 2) {
 		com_err(argv[0], 0, "Usage: ncheck <inode number> ...");
 		return;
@@ -74,19 +68,19 @@ void do_ncheck(int argc, char **argv)
 	if (check_fs_open(argv[0]))
 		return;
 
-	iw.iarray = malloc(sizeof(struct inode_info) * argc);
+	iw.iarray = malloc(sizeof(ext2_ino_t) * argc);
 	if (!iw.iarray) {
-		com_err("do_ncheck", ENOMEM,
+		com_err("ncheck", ENOMEM,
 			"while allocating inode info array");
 		return;
 	}
-	memset(iw.iarray, 0, sizeof(struct inode_info) * argc);
+	memset(iw.iarray, 0, sizeof(ext2_ino_t) * argc);
 
 	for (i=1; i < argc; i++) {
-		iw.iarray[i-1].ino = strtol(argv[i], &tmp, 0);
+		iw.iarray[i-1] = strtol(argv[i], &tmp, 0);
 		if (*tmp) {
 			com_err(argv[0], 0, "Bad inode - %s", argv[i]);
-			return;
+			goto error_out;
 		}
 	}
 
@@ -105,7 +99,8 @@ void do_ncheck(int argc, char **argv)
 		com_err("ncheck", retval, "while starting inode scan");
 		goto error_out;
 	}
-	
+
+	printf("Inode\tPathname\n");
 	while (ino) {
 		if (!inode.i_links_count)
 			goto next;
@@ -120,10 +115,17 @@ void do_ncheck(int argc, char **argv)
 			goto next;
 
 		iw.position = 0;
-		iw.parent = ino;
-		
+
+		retval = ext2fs_get_pathname(current_fs, ino, 0, &iw.parent);
+		if (retval) {
+			com_err("ncheck", retval, 
+				"while calling ext2fs_get_pathname");
+			goto next;
+		}
+
 		retval = ext2fs_dir_iterate(current_fs, ino, 0, 0,
 					    ncheck_proc, &iw);
+		ext2fs_free_mem(&iw.parent);
 		if (retval) {
 			com_err("ncheck", retval,
 				"while calling ext2_dir_iterate");
@@ -145,29 +147,6 @@ void do_ncheck(int argc, char **argv)
 		}
 	}
 
-	for (i=0, iinfo = iw.iarray; i < iw.num_inodes; i++, iinfo++) {
-		if (iinfo->parent == 0)
-			continue;
-		retval = ext2fs_get_pathname(current_fs, iinfo->parent,
-					     iinfo->ino, &iinfo->pathname);
-		if (retval)
-			com_err("ncheck", retval,
-				"while resolving pathname for inode %d (%d)",
-				iinfo->parent, iinfo->ino);
-	}
-	
-	printf("Inode\tPathname\n");
-	for (i=0, iinfo = iw.iarray; i < iw.num_inodes; i++, iinfo++) {
-		if (iinfo->parent == 0) {
-			printf("%u\t<inode not found>\n", iinfo->ino);
-			continue;
-		}
-		printf("%u\t%s\n", iinfo->ino, iinfo->pathname ?
-		       iinfo->pathname : "<unknown pathname>");
-		if (iinfo->pathname)
-			free(iinfo->pathname);
-	}
-
 error_out:
 	free(iw.iarray);
 	if (scan)
